<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <title>Toric Code Visualizer</title>

    <link rel="shortcut icon" href="{{ url_for('static', filename='favicon.ico') }}">
    <link rel="stylesheet" href="{{ url_for('static', filename='css/main.css') }}">

    <!-- <link rel="preconnect" href="https://fonts.googleapis.com">
    <link rel="preconnect" href="https://fonts.gstatic.com" crossorigin>
    <link href="https://fonts.googleapis.com/css2?family=Rubik:wght@300&display=swap" rel="stylesheet"> -->
    
</head>
<body>
    <h1>Quantum Code Visualizer</h1>

    <div id='menu'>
        <div class='menu-item'><a href='/2d'>2D Codes</a></div>
        <div class='menu-item'><a href='/3d'>3D Codes</a></div>
    </div>

    <footer><p>By <a href="http://artix41.github.io">Arthur Pesah</a> and <a href="#">Eric Huang</a>, 2022. Powered by ThreeJS.</p></footer>
</body>
</html> 