body {
    background-color: #102542;

    margin-top: 4em;
    
    font-family: 'Rubik', sans-serif;
    color:#ffebeb;
}

a {
    color: inherit; 
    text-decoration: inherit;
}

a:hover {
    color: #ffbcbc;
}

h1 {
    margin:auto;
    margin-bottom: 3em;
    text-align:center;
    /* color: #ffbcbc; */
    font-size: 4em;
    text-transform: uppercase;
}

#menu {
    margin: auto;
    width: 400px;
    height: 400px;
    font-size: 3em;
    text-align: center;
    text-transform: uppercase;

}


.menu-item {
    margin-bottom: 1.2em;
}


footer {
    position: fixed;
    left: 0;
    bottom: 0;
    width: 100%;
    /* color: #ffbcbc; */
    margin-bottom: 0.2em;
    font-size: 1.2em;
}

footer a {
    text-decoration: underline dotted;
    text-underline-offset: 3px;
}

footer p {
    text-align: right;
    padding-right: 2em;
}

#instructions {
    color: white;
    position: absolute;
    top: 140px;
    left: 40px;
    font-size: 20px;
    opacity: 0.6;
}

#closingCross {
    color: white;
    position: absolute;
    top: 115px;
    left: 15px;
    font-size: 20px;
}

#returnArrow {
    opacity: 1;
    color: white;
    font-size: 50px;
    position: absolute;
    top: 15px;
    left: 15px;
}